/*
 * File: AgentBody.cpp
 *
 * Description: AgentBody class implementation
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "AgentBody.h"
#include "LogLevels.h"
#include <ostream>
#include "SExpressionWriter.h"
#include "ParseTree.h"
#include "MathUtils.h"

using namespace std;

const AgentBody::AgentJointType AgentBody::JointType[] =
{
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,
    HINGE,

    UNIVERSAL,
    HINGE,
    HINGE,
    HINGE,
    UNIVERSAL,
    HINGE,
    UNIVERSAL,
    UNIVERSAL,
    HINGE,
    HINGE,
    HINGE,
    UNIVERSAL,
    HINGE,
    UNIVERSAL
};

const char *AgentBody::JointEffector[] =
{
    "he2",
    "he1",
    "rae1",
    "rae2",
    "rae3",
    "rae4",
    "rle1",
    "rle2",
    "rle3",
    "rle4",
    "rle5",
    "rle6",
    "lae1",
    "lae2",
    "lae3",
    "lae4",
    "lle1",
    "lle2",
    "lle3",
    "lle4",
    "lle5",
    "lle6",

    "rae1_2",
    "rae3",
    "rae4",
    "rle1",
    "rle2_3",
    "rle4",
    "rle5_6",
    "lae1_2",
    "lae3",
    "lae4",
    "lle1",
    "lle2_3",
    "lle4",
    "lle5_6"
};

const char *AgentBody::JointName[ILLEGAL_JOINT] =
{
    "NAO_HEAD_NECK",
    "NAO_NECK_TORSO",
    "NAO_RIGHT_TORSO_SHOULDER",
    "NAO_RIGHT_SHOULDER_UARM",
    "NAO_RIGHT_UARM_ELBOW",
    "NAO_RIGHT_ELBOW_LARM",
    "NAO_RIGHT_TORSO_HIP1",
    "NAO_RIGHT_HIP1_HIP2",
    "NAO_RIGHT_HIP2_THIGH",
    "NAO_RIGHT_THIGH_SHANK",
    "NAO_RIGHT_SHANK_ANKLE",
    "NAO_RIGHT_ANKLE_FOOT",
    "NAO_LEFT_TORSO_SHOULDER",
    "NAO_LEFT_SHOULDER_UARM",
    "NAO_LEFT_UARM_ELBOW",
    "NAO_LEFT_ELBOW_LARM",
    "NAO_LEFT_TORSO_HIP1",
    "NAO_LEFT_HIP1_HIP2",
    "NAO_LEFT_HIP2_THIGH",
    "NAO_LEFT_THIGH_SHANK",
    "NAO_LEFT_SHANK_ANKLE",
    "NAO_LEFT_ANKLE_FOOT",

    "RIGHT_ARM_BASE",
    "RIGHT_ARM_TOP",
    "RIGHT_ARM_MIDDLE",
    "RIGHT_LEG_BASE",
    "RIGHT_LEG_TOP",
    "RIGHT_LEG_MIDDLE",
    "RIGHT_LEG_FOOT",
    "LEFT_ARM_BASE",
    "LEFT_ARM_TOP",
    "LEFT_ARM_MIDDLE",
    "LEFT_LEG_BASE",
    "LEFT_LEG_TOP",
    "LEFT_LEG_MIDDLE",
    "LEFT_LEG_FOOT"
};

AgentBody::JointMap AgentBody::jointPerceptorMap;
AgentBody::JointMap AgentBody::jointNameMap;

AgentBody::AgentBody()
{
    if (jointPerceptorMap.empty())
    {
        jointPerceptorMap["hj2"]  = NAO_HEAD_NECK;
        jointPerceptorMap["hj1"]  = NAO_NECK_TORSO;
        jointPerceptorMap["raj1"] = NAO_RIGHT_TORSO_SHOULDER;
        jointPerceptorMap["raj2"] = NAO_RIGHT_SHOULDER_UARM;
        jointPerceptorMap["raj3"] = NAO_RIGHT_UARM_ELBOW;
        jointPerceptorMap["raj4"] = NAO_RIGHT_ELBOW_LARM;
        jointPerceptorMap["rlj1"] = NAO_RIGHT_TORSO_HIP1;
        jointPerceptorMap["rlj2"] = NAO_RIGHT_HIP1_HIP2;
        jointPerceptorMap["rlj3"] = NAO_RIGHT_HIP2_THIGH;
        jointPerceptorMap["rlj4"] = NAO_RIGHT_THIGH_SHANK;
        jointPerceptorMap["rlj5"] = NAO_RIGHT_SHANK_ANKLE;
        jointPerceptorMap["rlj6"] = NAO_RIGHT_ANKLE_FOOT;
        jointPerceptorMap["laj1"] = NAO_LEFT_TORSO_SHOULDER;
        jointPerceptorMap["laj2"] = NAO_LEFT_SHOULDER_UARM;
        jointPerceptorMap["laj3"] = NAO_LEFT_UARM_ELBOW;
        jointPerceptorMap["laj4"] = NAO_LEFT_ELBOW_LARM;
        jointPerceptorMap["llj1"] = NAO_LEFT_TORSO_HIP1;
        jointPerceptorMap["llj2"] = NAO_LEFT_HIP1_HIP2;
        jointPerceptorMap["llj3"] = NAO_LEFT_HIP2_THIGH;
        jointPerceptorMap["llj4"] = NAO_LEFT_THIGH_SHANK;
        jointPerceptorMap["llj5"] = NAO_LEFT_SHANK_ANKLE;
        jointPerceptorMap["llj6"] = NAO_LEFT_ANKLE_FOOT;

        jointPerceptorMap["raj1_2"] = RIGHT_ARM_BASE;
        jointPerceptorMap["raj3"] = RIGHT_ARM_TOP;
        jointPerceptorMap["raj4"] = RIGHT_ARM_MIDDLE;
        jointPerceptorMap["rlj1"] = RIGHT_LEG_BASE;
        jointPerceptorMap["rlj2_3"] = RIGHT_LEG_TOP;
        jointPerceptorMap["rlj4"] = RIGHT_LEG_MIDDLE;
        jointPerceptorMap["rlj5_6"] = RIGHT_LEG_FOOT;
        jointPerceptorMap["laj1_2"] = LEFT_ARM_BASE;
        jointPerceptorMap["laj3"] = LEFT_ARM_TOP;
        jointPerceptorMap["laj4"] = LEFT_ARM_MIDDLE;
        jointPerceptorMap["llj1"] = LEFT_LEG_BASE;
        jointPerceptorMap["llj2_3"] = LEFT_LEG_TOP;
        jointPerceptorMap["llj4"] = LEFT_LEG_MIDDLE;
        jointPerceptorMap["llj5_6"] = LEFT_LEG_FOOT;
    }

    if (jointNameMap.empty())
    {
        jointNameMap["NAO_HEAD_NECK"] = NAO_HEAD_NECK;
        jointNameMap["NAO_NECK_TORSO"] = NAO_NECK_TORSO;
        jointNameMap["NAO_RIGHT_TORSO_SHOULDER"] = NAO_RIGHT_TORSO_SHOULDER;
        jointNameMap["NAO_RIGHT_SHOULDER_UARM"] = NAO_RIGHT_SHOULDER_UARM;
        jointNameMap["NAO_RIGHT_UARM_ELBOW"] = NAO_RIGHT_UARM_ELBOW;
        jointNameMap["NAO_RIGHT_ELBOW_LARM"] = NAO_RIGHT_ELBOW_LARM;
        jointNameMap["NAO_RIGHT_TORSO_HIP1"] = NAO_RIGHT_TORSO_HIP1;
        jointNameMap["NAO_RIGHT_HIP1_HIP2"] = NAO_RIGHT_HIP1_HIP2;
        jointNameMap["NAO_RIGHT_HIP2_THIGH"] = NAO_RIGHT_HIP2_THIGH;
        jointNameMap["NAO_RIGHT_THIGH_SHANK"] = NAO_RIGHT_THIGH_SHANK;
        jointNameMap["NAO_RIGHT_SHANK_ANKLE"] = NAO_RIGHT_SHANK_ANKLE;
        jointNameMap["NAO_RIGHT_ANKLE_FOOT"] = NAO_RIGHT_ANKLE_FOOT;
        jointNameMap["NAO_LEFT_TORSO_SHOULDER"] = NAO_LEFT_TORSO_SHOULDER;
        jointNameMap["NAO_LEFT_SHOULDER_UARM"] = NAO_LEFT_SHOULDER_UARM;
        jointNameMap["NAO_LEFT_UARM_ELBOW"] = NAO_LEFT_UARM_ELBOW;
        jointNameMap["NAO_LEFT_ELBOW_LARM"] = NAO_LEFT_ELBOW_LARM;
        jointNameMap["NAO_LEFT_TORSO_HIP1"] = NAO_LEFT_TORSO_HIP1;
        jointNameMap["NAO_LEFT_HIP1_HIP2"] = NAO_LEFT_HIP1_HIP2;
        jointNameMap["NAO_LEFT_HIP2_THIGH"] = NAO_LEFT_HIP2_THIGH;
        jointNameMap["NAO_LEFT_THIGH_SHANK"] = NAO_LEFT_THIGH_SHANK;
        jointNameMap["NAO_LEFT_SHANK_ANKLE"] = NAO_LEFT_SHANK_ANKLE;
        jointNameMap["NAO_LEFT_ANKLE_FOOT"] = NAO_LEFT_ANKLE_FOOT;

        jointNameMap["RIGHT_ARM_BASE"] = RIGHT_ARM_BASE;
        jointNameMap["RIGHT_ARM_TOP"] = RIGHT_ARM_TOP;
        jointNameMap["RIGHT_ARM_MIDDLE"] = RIGHT_ARM_MIDDLE;
        jointNameMap["RIGHT_LEG_BASE"] = RIGHT_LEG_BASE;
        jointNameMap["RIGHT_LEG_TOP"] = RIGHT_LEG_TOP;
        jointNameMap["RIGHT_LEG_MIDDLE"] = RIGHT_LEG_MIDDLE;
        jointNameMap["RIGHT_LEG_FOOT"] = RIGHT_LEG_FOOT;
        jointNameMap["LEFT_ARM_BASE"] = LEFT_ARM_BASE;
        jointNameMap["LEFT_ARM_TOP"] = LEFT_ARM_TOP;
        jointNameMap["LEFT_ARM_MIDDLE"] = LEFT_ARM_MIDDLE;
        jointNameMap["LEFT_LEG_BASE"] = LEFT_LEG_BASE;
        jointNameMap["LEFT_LEG_TOP"] = LEFT_LEG_TOP;
        jointNameMap["LEFT_LEG_MIDDLE"] = LEFT_LEG_MIDDLE;
        jointNameMap["LEFT_LEG_FOOT"] = LEFT_LEG_FOOT;
        jointNameMap["TIME_DELAY"] = ILLEGAL_JOINT;
    }
}

AgentBody::~AgentBody()
{
}

void AgentBody::SetJointInfo(string jointPercept, float dt, float ax1, float ax2)
{
    JointMap::const_iterator jointptr = jointPerceptorMap.find(jointPercept);
    if (jointptr==jointPerceptorMap.end())
        Log(LOG_ERROR) << "Unknown joint percept received: " << jointPercept;
    else
    {
        Joint joint = jointptr->second;
        Log(LOG_DEBUG_SENSE) << JointName[joint] << " ("<< ax1 << ' '<< ax2 << ')';
        joints[joint].SetJointAxis(ax1, ax2, dt);
    }
}

void AgentBody::SetGyroRate(string gyroName, Vector3D gyro)
{
    gyroRate = gyro;
    Log(LOG_DEBUG_SENSE) << "Gyro " << gyroName << ": " << gyroRate;
}

void AgentBody::SetForceResistanceInfo(string frpName, Vector3D center, Vector3D force)
{
    AgentFoot foot = (frpName == "lf") ? LEFT_FOOT : RIGHT_FOOT;
    footTouchForceInfo[foot].center = center;
    footTouchForceInfo[foot].force = force;

    Log(LOG_DEBUG_SENSE) << "FRP " << frpName << ": center->" << center
                        << " force->" << force;
}

void AgentBody::SaveAgentState(ostream &os, string stateName)
{
    SExpressionWriter writer(os, "State");
    writer.BeginSub("name").Add(stateName).EndSub();
    writer.AddVector3D("Gyro", gyroRate).Endl();

    writer.BeginSub("LeftFootForce");
    writer.AddVector3D("Center", footTouchForceInfo[LEFT_FOOT].center);
    writer.AddVector3D("Force", footTouchForceInfo[LEFT_FOOT].force);
    writer.EndSub().Endl();

    writer.BeginSub("RightFootForce");
    writer.AddVector3D("Center", footTouchForceInfo[RIGHT_FOOT].center);
    writer.AddVector3D("Force", footTouchForceInfo[RIGHT_FOOT].force);
    writer.EndSub().Endl();

    for (int i = 0; i < ILLEGAL_JOINT; ++i)
    {
        writer.BeginSub(JointName[i]);
        writer.BeginSub("axis1").Add(joints[i].Axis(AX1)).EndSub();
        if (JointType[i] == UNIVERSAL)
            writer.BeginSub("axis2").Add(joints[i].Axis(AX2)).EndSub();
        writer.BeginSub("rate1").Add(joints[i].Rate(AX1)).EndSub();
        if (JointType[i] == UNIVERSAL)
            writer.BeginSub("rate2").Add(joints[i].Rate(AX2)).EndSub();
        writer.EndSub().Endl();
    }
}

string AgentBody::ReadState( ParseTree *state )
{
    string name = "none";
    Log(LOG_DEBUG_STREADER) << "Reading new state...";
    for (unsigned i = 0; i < state->GetNumOfKeys(); ++i)
    {
        if (state->GetKey(i) == "name")
            name = state->GetData(i)->GetKey(0);
        else if (state->GetKey(i) == "Gyro")
            state->GetData(i)->GetVector(gyroRate);
        else if (state->GetKey(i) == "LeftFootForce")
            ReadFootForce(LEFT_FOOT, state->GetData(i));
        else if (state->GetKey(i) == "RightFootForce")
            ReadFootForce(RIGHT_FOOT, state->GetData(i));
        else
            joints[joint(state->GetKey(i))].ReadJointInfo(state->GetData(i));
    }
    Log(LOG_DEBUG_STREADER) << "Reading finished for state: " << name;

    return name;
}

AgentBody::Joint AgentBody::joint(string name)
{
    JointMap::const_iterator jointptr = jointNameMap.find(name);
    if (jointptr != jointNameMap.end())
        return jointptr->second;
    else
    {
        Log(LOG_ERROR) << "Joint not found: " << name;
        return ILLEGAL_JOINT;
    }
}

void AgentBody::ReadFootForce( AgentBody::AgentFoot foot, ParseTree *frpInfo )
{
    for (unsigned i = 0; i < frpInfo->GetNumOfKeys(); ++i)
    {
        if (frpInfo->GetKey(i) == "Center")
            frpInfo->GetData(i)->GetVector(footTouchForceInfo[foot].center);
        else if (frpInfo->GetKey(i) == "Force")
            frpInfo->GetData(i)->GetVector(footTouchForceInfo[foot].force);
        else
            Log(LOG_ERROR) << "Unknown tag in FRP " << frpInfo->GetKey(i);
    }
    Log(LOG_DEBUG_STREADER) << "\tReading FRP for "<< (foot == LEFT_FOOT ? "left"
            : "right" )<< " foot: "<< footTouchForceInfo[foot].center
            << footTouchForceInfo[foot].force;
}

void AgentBody::JointInfo::SetJointAxis(AngDeg ax1, AngDeg ax2, float dt)
{
    if (dt > 0)
    {
        rate[AX1] = gDegToRad(gNormalizeDeg(ax1 - axis[AX1])) / dt;
        rate[AX2] = gDegToRad(gNormalizeDeg(ax2 - axis[AX2])) / dt;
    }
    axis[AX1] = ax1;
    axis[AX2] = ax2;
}

void AgentBody::JointInfo::SetSentRates(AngDeg rt1, AngDeg rt2)
{
    sentRate[AX1] = rt1;
    sentRate[AX2] = rt2;
}

void AgentBody::JointInfo::ReadJointInfo( ParseTree *jointInfo )
{
    for (unsigned i = 0; i < jointInfo->GetNumOfKeys(); ++i)
    {
        if (jointInfo->GetKey(i) == "axis1")
            jointInfo->GetData(i)->GetKey(0, axis[AX1]);
        else if (jointInfo->GetKey(i) == "axis2")
            jointInfo->GetData(i)->GetKey(0, axis[AX2]);
        else if (jointInfo->GetKey(i) == "rate1")
            jointInfo->GetData(i)->GetKey(0, rate[AX1]);
        else if (jointInfo->GetKey(i) == "rate2")
            jointInfo->GetData(i)->GetKey(0, rate[AX2]);
        else
            Log(LOG_DEBUG_STREADER) << "Unknown joint info tag: " << jointInfo->GetKey(i);
    }
    Log(LOG_DEBUG_STREADER) << "\tReading joint info: "<< axis[AX1] << ' '<< axis[AX2] << ' '
            << rate[AX1] << ' '<< rate[AX2];
}
