/*
 * File: AgentBody.h
 *
 * Description: Agent body information class
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef AGENTBODY_H_
#define AGENTBODY_H_

#include <string>
#include <map>
#include "Vector.h"
#include "LogLevels.h"
#include <iosfwd>

using namespace std;

class ParseTree;

class AgentBody
{
    public:
        enum Joint
        {
            NAO_HEAD_NECK,
            NAO_NECK_TORSO,
            NAO_RIGHT_TORSO_SHOULDER,
            NAO_RIGHT_SHOULDER_UARM,
            NAO_RIGHT_UARM_ELBOW,
            NAO_RIGHT_ELBOW_LARM,
            NAO_RIGHT_TORSO_HIP1,
            NAO_RIGHT_HIP1_HIP2,
            NAO_RIGHT_HIP2_THIGH,
            NAO_RIGHT_THIGH_SHANK,
            NAO_RIGHT_SHANK_ANKLE,
            NAO_RIGHT_ANKLE_FOOT,
            NAO_LEFT_TORSO_SHOULDER,
            NAO_LEFT_SHOULDER_UARM,
            NAO_LEFT_UARM_ELBOW,
            NAO_LEFT_ELBOW_LARM,
            NAO_LEFT_TORSO_HIP1,
            NAO_LEFT_HIP1_HIP2,
            NAO_LEFT_HIP2_THIGH,
            NAO_LEFT_THIGH_SHANK,
            NAO_LEFT_SHANK_ANKLE,
            NAO_LEFT_ANKLE_FOOT,

            RIGHT_ARM_BASE,
            RIGHT_ARM_TOP,
            RIGHT_ARM_MIDDLE,
            RIGHT_LEG_BASE,
            RIGHT_LEG_TOP,
            RIGHT_LEG_MIDDLE,
            RIGHT_LEG_FOOT,
            LEFT_ARM_BASE,
            LEFT_ARM_TOP,
            LEFT_ARM_MIDDLE,
            LEFT_LEG_BASE,
            LEFT_LEG_TOP,
            LEFT_LEG_MIDDLE,
            LEFT_LEG_FOOT,
            ILLEGAL_JOINT
        };

        enum AgentJointType
        {
            UNIVERSAL,
            HINGE
        };

        enum AgentFoot
        {
            LEFT_FOOT,
            RIGHT_FOOT
        };

        enum JointAxis
        {
            AX1,
            AX2
        };

        class JointInfo
        {
            public:
                JointInfo()
                {
                    axis[0] = axis[1] = 0;
                    rate[0] = rate[1] = 0;
                    sentRate[0] = sentRate[1] = 0;
                }

                AngDeg Axis(JointAxis ax) const { return axis[ax]; }
                AngDeg Rate(JointAxis ax) const { return rate[ax]; }
                AngDeg SentRate(JointAxis ax) const { return sentRate[ax]; }
                void SetJointAxis(AngDeg ax1, AngDeg ax2 = 0, float dt = 0);
                void SetSentRates(AngDeg rt1, AngDeg rt2 = 0);
                void ReadJointInfo( ParseTree *jointInfo );

            private:
                AngDeg axis[2];
                AngDeg sentRate[2];
                AngDeg rate[2];
        };

        struct ForceResistanceInfo
        {
            Vector3D center;
            Vector3D force;
            void Reset()
            {
                center.Zero();
                force.Zero();
            }
            bool IsAny() const
            {
                return !force.IsZero();
            }
        };

        typedef map<string, AgentBody::Joint> JointMap;

    public:
        static const AgentJointType JointType[];
        static const char *JointEffector[];
        static const char *JointName[ILLEGAL_JOINT];
        static const int CurrentFinalJoint = RIGHT_ARM_BASE;

        AgentBody();
        ~AgentBody();

        static Joint joint( string name );
        void SetJointInfo(string jointPercept, float dt, float ax1, float ax2 = 0);
        void SetGyroRate(string gyroName, Vector3D gyro);
        void SetForceResistanceInfo(string frpName, Vector3D center, Vector3D force);
        void SaveAgentState(ostream &os, string stateName);
        string ReadState( ParseTree *state );

        void ClearFRPData()
        {
            footTouchForceInfo[0].Reset();
            footTouchForceInfo[1].Reset();
        }

        const JointInfo &JointData(Joint joint) const
        {
            return joints[joint];
        }

        JointInfo &JointData(Joint joint)
        {
            return joints[joint];
        }

        const Vector3D &GyroRate() const
        {
            return gyroRate;
        }

        const ForceResistanceInfo &FootTouchForceInfo(AgentFoot foot) const
        {
            return footTouchForceInfo[foot];
        }

    private:
        static JointMap jointPerceptorMap;
        static JointMap jointNameMap;

        JointInfo joints[ILLEGAL_JOINT];
        Vector3D gyroRate;
        ForceResistanceInfo footTouchForceInfo[2];

        void ReadFootForce( AgentFoot foot, ParseTree *frpInfo );
};

#endif /*AGENTBODY_H_*/
