/*
 * File: BaseAgent.cpp
 *
 * Description: BaseAgent implementation for agent's sense & act
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "BaseAgent.h"
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_array.hpp>
#include <fstream>
#include <string>
#include <ruby.h>
#include "undef_package.h"
#include "LogLevels.h"
#include "MathUtils.h"
#include "TransitionManager.h"
#include "AgentBodyStates.h"
#include "ActionManager.h"

using namespace std;
using namespace boost;

BaseAgent::BaseAgent() :
    wm(&connection), nextAction(ActionManager::ILLEGAL_ACT), acts(wm),
    newStep(true), isInFinalCycle(false)
{
}

BaseAgent::~BaseAgent()
{
}

BaseAgent::RubyException::RubyException(): Exception()
{
    VALUE ret = rb_inspect(rb_gv_get("$!"));
    msg = RSTRING(ret)->ptr;
}

void BaseAgent::Connect(rcss::net::Addr addr)
{
    connection.Connect(addr);
}

inline VALUE BaseAgent::RubyRun(const char *command)
{
    int error;
    VALUE res = rb_eval_string_protect(command,&error);
    ThrowIf<RubyException>(error);
    return res;
}

void BaseAgent::InitRubyFunctions()
{
    ifstream scriptFile("functions.rb");
    if (scriptFile)
    {
        ruby_init();
        scriptFile.seekg(0, ios::end);
        unsigned size = scriptFile.tellg();
        boost::scoped_array<char> buffer(new char[size+1]);

        scriptFile.seekg(0);
        scriptFile.read(buffer.get(), size);
        ThrowIf(!scriptFile, "File read error: functions.rb");
        buffer[size] = 0;

        RubyRun(buffer.get());

        Log(LOG_DEBUG) << "Ruby functions initialized successfully";
    }
    else
        Log(LOG_DEBUG) << "No ruby functions initialized";
}

inline void BaseAgent::ConstructAgent()
{
    string agent = "(scene rsg/agent/nao/nao.rsg)";
    connection.Send(agent);
}

inline void BaseAgent::Init()
{
    connection.Send("(init (unum 0)(teamname "+wm.TeamName()+"))");
}

void BaseAgent::Run(bool enableUserCommand)
{
    thread senseThread(bind(&WorldModelCore::Sense, &wm));

    ConstructAgent();
    wm.WaitForNewSense();
    Init();

    if (enableUserCommand)
        thread userInputThread(bind(&BaseAgent::UserCommandManager, this));
    thread actThread(bind(&BaseAgent::ActManager, this));

    while (wm.IsActive())
    {
//		Log(LOG_DEBUG)<<"\nWaiting for sense info...";
        wm.WaitForNewSense();
//		Log(LOG_DEBUG)<<"fin!";
        // NOTICE: it is needed when not using the timer!
//		if (wm.IsNewSenseInfo())
//		{
            Log(LOG_DEBUG)<<"Thinking";
            Think();
//		}
    }
    actThread.join();
}

void BaseAgent::SendCommands()
{
    if(!commands.str().empty())
    {
        connection.Send(commands.str());
        commands.str("");
    }
}

void BaseAgent::Beam(float x, float y, AngDeg zrotate)
{
    commands << setiosflags(ios::fixed) << setprecision(2) << "(beam " << x << ' ' << y << ' ' << zrotate << ')';
}

void BaseAgent::SetJointVelocity(AgentBody::Joint joint, float vel1, float vel2)
{
    commands << '(' << AgentBody::JointEffector[joint] << ' ' << vel1;
    if (AgentBody::JointType[joint] == AgentBody::UNIVERSAL)
        commands << ' ' << vel2;
    else if ( vel2 != 0 )
        Log(LOG_ERROR) << "The joint " << AgentBody::JointEffector[joint] << " needs only 1 parameter!";
    commands << ')';
}

void BaseAgent::LoadDataFile(const char *filename)
{
    Log(LOG_DEBUG) << "Loading data file: " << filename;
    acts.LoadFile(filename);
}

void BaseAgent::LoadFilesFromList(string listFileName)
{
    ifstream listfile(listFileName.c_str());
    if (!listfile)
    {
        Log(LOG_ERROR) << "Cannot open file: " << listFileName;
        return;
    }

    string filename;
    getline(listfile, filename);
    while (listfile)
    {
        LoadDataFile(filename.c_str());
        getline(listfile, filename);
    }
}

void BaseAgent::ActManager()
{
    Log(LOG_ACTION_MANAGER)<<"Starting Action Manager...";
    try
    {
        InitRubyFunctions();
        while (wm.IsActive())
        {
            // TODO: may be replaced by a timer, specially if there is async sense and act
            wm.WaitForNewSense();

            Log(LOG_DEBUG)<<"Acting!";
            if (nextAction != ActionManager::ILLEGAL_ACT)
            {
                acts.SetNextAction(nextAction);
                mutex::scoped_lock lock(actMutex);
                nextAction = ActionManager::ILLEGAL_ACT;
            }
            Act();

            SendCommands();
        }

    }
    catch (std::exception &e)
    {
        Log(LOG_ERROR) << "A bad thing happened in the ActManager: " << e.what();
    }
    catch (...)
    {
        Log(LOG_ERROR) << "A bad thing happened in the ActManager :( ";
    }

    Log(LOG_ACTION_MANAGER)<<"Action Manager finished!";
}

void BaseAgent::SetAction( int act )
{
    mutex::scoped_lock lock(actMutex);
    nextAction = act;
}

void BaseAgent::SetActionType( int type, const char *actName )
{
    acts.SetActionType(type, actName);
}

void BaseAgent::Act()
{
    Log(LOG_JOINT_CONTROLLER) << FUNCTION_LOG_HEADER;
    if (!acts.IsActing())
    {
         Log(LOG_JOINT_CONTROLLER) << "Nothing to be done!!!";
        return;
    }

//    Vector3D gyro = wm.GyroRate();
//    AgentBody::ForceResistanceInfo leftFoot = wm.FootTouchForceInfo(AgentBody::LEFT_FOOT);
//    AgentBody::ForceResistanceInfo rightFoot = wm.FootTouchForceInfo(AgentBody::RIGHT_FOOT);
//    Log(LOG_FRPCHECK) << "Gyro: " << gyro << "\nleft FRP: center->" << leftFoot.center
//                    << " force->" << leftFoot.force << "\nright FRP: center->" << rightFoot.center
//                    << " force->" << rightFoot.force;
    isStepCompleted = true;

    acts.UpdateParameters();
    TransitionData::JointTransitionVector &activeJoints = acts.JointActions();
    for(unsigned i=0; i<activeJoints.size(); ++i)
    {
        if (activeJoints[i].joint!=AgentBody::ILLEGAL_JOINT)
        {
            float rate1 = GetNewJointRate(activeJoints[i], AgentBody::AX1);
            float rate2 =
                (AgentBody::JointType[activeJoints[i].joint] == AgentBody::HINGE) ? 0
                        : GetNewJointRate(activeJoints[i], AgentBody::AX2);
            Log(LOG_JOINT_CONTROLLER) << "Sending joint command: "
                                << AgentBody::JointName[activeJoints[i].joint]
                                << ' ' << rate1 << ' ' << rate2 << '\n';

            SetJointVelocity( activeJoints[i].joint, rate1, rate2 );
        }
        else{
            Log(LOG_JOINT_CONTROLLER) << "Start Time: " << activeJoints[i].delay << " spent simtime:" <<
                    wm.GetSpentTime(activeJoints[i].delay) << " Delay: " <<activeJoints[i].delay;
            if (wm.GetSpentTime(acts.StepStartTime()) < activeJoints[i].delay)
                isStepCompleted = false;
        }
    }

    newStep = false;
    if (isStepCompleted)
    {
        if (!acts.CurrentTransition().CheckFunction().empty())
        {
            Log(LOG_JOINT_CONTROLLER) << "Calling ruby function";
            VALUE res = RubyRun(acts.CurrentTransition().CheckFunction().c_str());
            float val = NUM2DBL(res);
            if (val < 0.6)
                return;
        }
        acts.GotoNextStep();
        isInFinalCycle = false;
        newStep = true;
    }
}

inline float BaseAgent::GetNewJointRate(
        const TransitionData::JointTransitionData &jointData,
        AgentBody::JointAxis ax)
{
    const float SERVER_STEP = 0.02;
    const float HIGH_PRECISION = gDegToRad(15.0F) / SERVER_STEP;
    float newRate = 0;
    float currentRate = wm.JointInformation(jointData.joint).Rate(ax);
    AngDeg jointax = wm.JointAxis(jointData.joint, ax);
    float diff = gNormalizeRad(gDegToRad(jointData.targetax[ax] - jointax)
            - currentRate * SERVER_STEP) / SERVER_STEP;
    int diffSign = sign(diff);

//    Log(LOG_JOINT_CONTROLLER) << "Axis: " << ax << " rt: " << currentRate << " diff: "
//                << diff * SERVER_STEP << " HP: " << HIGH_PRECISION * SERVER_STEP
//                << " angle: " << jointax << " target: " << jointData.targetax[ax]
//                << " rate: " << jointData.rate[ax];

    if (newStep)
        JointRateSigns[jointData.joint][ax] = diffSign;

    if (fabs(jointData.targetax[ax]) >= 999 )
    {
        diffSign = sign(jointData.targetax[ax]);
        diff = jointData.targetax[ax];
    }
    else if (diffSign == JointRateSigns[jointData.joint][ax])
    {
        //NOTICE: diff is devided by SERVER_STEP
        if (fabs(diff) > 0.02 && jointData.rate[ax] > 0.01)
        {
            isStepCompleted = false;
            Log(LOG_JOINT_CONTROLLER) << "dist  Axis: " << ax << " rt: " << currentRate << " diff: "
                        << diff * SERVER_STEP << " HP: " << HIGH_PRECISION * SERVER_STEP
                        << " angle: " << jointax << " target: " << jointData.targetax[ax]
                        << " rate: " << jointData.rate[ax];
        }
    }
    else if (fabs(diff) > HIGH_PRECISION && jointData.rate[ax] > 0.01)
    {
        JointRateSigns[jointData.joint][ax] = diffSign;
        isStepCompleted = false;
        Log(LOG_JOINT_CONTROLLER) << "prec!";
    }
    else
        return 0;

    newRate = diffSign * min(jointData.rate[ax], fabs(diff));

    float growRate = jointData.rate[ax] * 0.1;
    if (fabs(newRate) > fabs(currentRate) + growRate)
    {
        newRate = currentRate + diffSign * growRate;
    }
//    else if (fabs(newRate) < fabs(currentRate) - growRate)
//    {
//        newRate = currentRate - diffSign * growRate;
//    }

    return newRate;
}

float BaseAgent::GetNewJointRate2(
        const TransitionData::JointTransitionData &jointData,
        AgentBody::JointAxis ax)
{
    //  Log(LOG_JOINT_CONTROLLER)<<"**************new rate file";
    //TODO you should be careful, if you come to reduce speed you should never leave it! because it seems dangerous work
    const int maxStepToFinish = 5;//stepToFinsish step remained to reach final Point, be careful it's value should be >1 or <-1
    const float SERVER_STEP = 0.02;
    const float HIGH_PRECISION = gDegToRad(15.0F) / SERVER_STEP;
    float newRate = 0;
    float currentRate = wm.JointInformation(jointData.joint).Rate(ax);
    AngDeg jointax = wm.JointAxis(jointData.joint, ax);
    float diff = gNormalizeRad(gDegToRad(jointData.targetax[ax] - jointax)
            - currentRate * SERVER_STEP) / SERVER_STEP;
    int diffSign = sign(diff);
    /*  Log(LOG_JOINT_CONTROLLER) <<"Axis: "<< ax<< " rt: "
     << currentRate<< " diff: "<< diff * SERVER_STEP << " HP: "
     << HIGH_PRECISION* SERVER_STEP<< " angle: "<< jointax
     << " target: "<< jointData.targetax[ax]<< " rate: "
     << jointData.rate[ax];
     */

    if (newStep)
        JointRateSigns[jointData.joint][ax] = diffSign;

    if (fabs(jointData.targetax[ax]) >= 999)
    {
        diffSign = sign(jointData.targetax[ax]);
        diff = jointData.targetax[ax];
    } else if (diffSign == JointRateSigns[jointData.joint][ax])
    {
        //NOTICE: diff is devided by SERVER_STEP
        if (fabs(diff) > 0.02&& jointData.rate[ax] > 0.01)
        {
            isStepCompleted = false;
            //          Log(LOG_JOINT_CONTROLLER) << "dist  Axis: "<< AgentBody::JointName[ax] << " rt: "
            //                  << currentRate << " diff: "<< diff * SERVER_STEP << " HP: "
            //                  << HIGH_PRECISION * SERVER_STEP<< " angle: "<< jointax
            //                  << " target: "<< jointData.targetax[ax]<< " rate: "
            //                  << jointData.rate[ax];
        }
    } else if (fabs(diff) > HIGH_PRECISION && jointData.rate[ax] > 0.01)
    {
        JointRateSigns[jointData.joint][ax] = diffSign;
        isStepCompleted = false;
        Log(LOG_JOINT_CONTROLLER) << "prec!";
    } else
        return 0;

    newRate = diffSign * min(jointData.rate[ax], fabs(diff));

    float growRate = jointData.rate[ax] * 0.1;
    if (fabs(newRate) > fabs(currentRate) + growRate)
    {
        newRate = currentRate + diffSign * growRate;
    }
    //    else if (fabs(newRate) < fabs(currentRate) - growRate)
    //    {
    //        newRate = currentRate - diffSign * growRate;
    //    }
    float jointMove = 0;//= currentRate * SERVER_STEP;
    Log(LOG_JOINT_CONTROLLER)
            <<"&&&&&&&&&&&&&&&&&&&&&&&&& in speed function &&&&&&&&&&&&&&&&&&&&&&&&&";
    const float reducedRate = jointData.rate[ax] * (1.0
            / (float)maxStepToFinish);
    float simCurrentRate = currentRate;
    Log(LOG_JOINT_CONTROLLER)<<"currentRate: "<<currentRate<<" reducedRate: "
            <<reducedRate<<" and jonitRate[ax] is: "<<jointData.rate[ax];
    Log(LOG_JOINT_CONTROLLER)<<"**before loop! **";
    for (int i = 0; i <= maxStepToFinish; ++i)//predict position after 5 cycle by current Rate!
    {
        jointMove += simCurrentRate * SERVER_STEP;
        simCurrentRate -= sign(currentRate) * reducedRate;
        Log(LOG_JOINT_CONTROLLER)<<"i "<<i<<" jointMove "<<gRadToDeg(jointMove)
                <<" simCurrentRate "<<simCurrentRate;
        if (sign(simCurrentRate) != sign(currentRate) )
            break;
    }
    Log(LOG_JOINT_CONTROLLER)<<"**end of loop! **";
    AngDeg difToTarget = fabs(jointData.targetax[ax] - jointax);//gRadToDeg(fabs( diff * SERVER_STEP ));
    jointMove = gRadToDeg(fabs(jointMove));
    Log(LOG_JOINT_CONTROLLER)<<" current jointax : "<<jointax
            <<" and target is "<<jointData.targetax[ax];
    Log(LOG_JOINT_CONTROLLER)<<"difToTarget "<<difToTarget
            <<" and jointMove is : "<<jointMove;
    if (((jointMove >= difToTarget && currentRate) || isInFinalCycle)
            && jointMove)//reduce speed!
    {
        if (!newRate)
            return newRate;
        Log(LOG_JOINT_CONTROLLER)
                <<"######################### i wana reduce speed ######################### ";
        isInFinalCycle = true;
        Log(LOG_JOINT_CONTROLLER)<<"before reducing speed newRate =  "<<newRate
                <<" and currentRate : "<<currentRate;
        //      newRate = currentRate - diffSign * reducedRate;
        growRate = sign(currentRate) * reducedRate;
        const float lastNewRate = newRate;
        newRate = currentRate - growRate;
        if (sign(newRate) != sign(currentRate))
        {
            Log(LOG_JOINT_CONTROLLER)<<"ops! error!";
            if (sign(jointax - jointData.targetax[ax]) != sign(currentRate))//agent was cross over target if this condition happened
                //TODO this condition is not corrent for zero targets! or somthing else is wrong for zero target!
                newRate = 0;
            else
                newRate = lastNewRate;
        }
        //      newRate = max(0.0F, fabs(newRate));
        //      if(JointRateSigns[jointData.joint][ax] != sign(newRate))// && newRate > currentRate)
        //      {
        //          /*cross the jointTarget!*/
        //          /*if(diffSign == sign(currentRate))
        //              newRate = 0;
        //          else*/
        //              newRate = currentRate;
        //      }

        Log(LOG_JOINT_CONTROLLER)<<"growRate is : "<<growRate
                <<" and newRate is : "<<newRate;
    } else
    {
        //       isInFinalCycle = false;
    }
    return newRate;
}

void BaseAgent::CheckStability()
{
    if (!isStepCompleted)
        return;
//	Log(LOG_FRPCHECK) << FUNCTION_LOG_HEADER;
    Vector3D gyro = wm.GyroRate();
    AgentBody::ForceResistanceInfo leftFoot = wm.FootTouchForceInfo(AgentBody::LEFT_FOOT);
    AgentBody::ForceResistanceInfo rightFoot = wm.FootTouchForceInfo(AgentBody::RIGHT_FOOT);
//    Log(LOG_FRPCHECK) << "Gyro: " << gyro << "\nleft FRP: center->" << leftFoot.center
//                    << " force->" << leftFoot.force << "\nright FRP: center->" << rightFoot.center
//                    << " force->" << rightFoot.force;
//    if (fabs(gyro.y()) > 5.0)
//    {
//        Log(LOG_FRPCHECK) << "Gyro Alarming... " << gyro;
//        isStepCompleted = false;
//    }
    if (fabs(gyro.x()) > 5.0)
    {
        Log(LOG_FRPCHECK) << "X Gyro Alarming... " << gyro;
        isStepCompleted = false;
    }
    bool frpOK = false;
    if (leftFoot.center.Length() < 0.7 && leftFoot.force.z() > leftFoot.force.x()
            + leftFoot.force.y() + 4.0 && rightFoot.force.x() + rightFoot.force.y() < 1)
        frpOK = true;
    if (rightFoot.center.Length() < 0.7 && rightFoot.force.z() > rightFoot.force.x()
            + rightFoot.force.y() + 4.0 && rightFoot.force.x() + rightFoot.force.y() < 1)
        frpOK = true;
    if (!frpOK)
    {
//        Log(LOG_FRPCHECK) << "FRP Alarming...";
//        isStepCompleted = false;
//        Log(LOG_FRPCHECK) << "Gyro: " << gyro << "\nleft FRP: center->" << leftFoot.center
//                        << " force->" << leftFoot.force << "\nright FRP: center->" << rightFoot.center
//                        << " force->" << rightFoot.force;
    }
}

void BaseAgent::ResetAgent(float x, float y, AngDeg zrotate)
{
    connection.Reconnect();
    ConstructAgent();
    wm.WaitForNewSense();
    Init();
    Beam(x,y,zrotate);
}
