/*
 * File: BaseAgent.h
 *
 * Description: BaseAgent provides basic agent functionality (sense & act).
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef BASEAGENT_H_
#define BASEAGENT_H_

#include <sstream>
#include "rcssnet/addr.hpp"
#include <vector>
#include <string>
#include <boost/thread/mutex.hpp>
#include "ConnectionManager.h"
#include "WorldModel.h"
#include "AgentBody.h"
#include "ActionManager.h"
#include "2DShapes.h"
#include "TrainerControl.h"

using namespace std;

typedef unsigned long VALUE;

class BaseAgent
{
    public:
        class RubyException: public Exception
        {
            public:
                RubyException();
        };

    public:
        BaseAgent();
        virtual ~BaseAgent();
        void Connect(rcss::net::Addr addr);
        void Run(bool enableUserCommand = false);
        void LoadDataFile(const char *filename);
        void LoadFilesFromList(string listFileName);

    protected:
        WorldModel wm;
        TrainerControl trainer;

        void Beam(float x, float y, AngDeg zrotate);
        void SetJointVelocity(AgentBody::Joint joint, float vel1, float vel2 = 0.0);
        void SetAction( int act );
        void SetActionType( int type, const char *actName );
        void ResetAgent(float x, float y, AngDeg zrotate);

        int NextAction() const
        {
            return nextAction;
        }

        void StopAction()
        {
            acts.Stop();
            nextAction = ActionManager::ILLEGAL_ACT;
        }

        bool IsActing()
        {
            return acts.IsActing();
        }

        void SetTransitionParameter(const char *sourceStateName,
                                    const char *destStateName, const char *name, float val)
        {
            acts.SetTransitionParameter(sourceStateName, destStateName, name, val);
        }

    private:
        enum Part {
            NECK,
            LEFT_HAND,
            RIGHT_HAND,
            LEFT_LEG,
            RIGHT_LEG
        };

        int nextAction;
        ConnectionManager connection;
        stringstream commands;
        boost::mutex actMutex;
        ActionManager acts;
        bool newStep;
        bool isStepCompleted;
        bool isInFinalCycle;
        int JointRateSigns[AgentBody::ILLEGAL_JOINT][2];

        VALUE RubyRun(const char *command);
        void InitRubyFunctions();
        void ConstructAgent();
        void Init();
        void SendCommands();
        void ActManager();
        void UserCommandManager();
        void Act();
        float GetNewJointRate(
                const TransitionData::JointTransitionData &jointData,
                AgentBody::JointAxis ax);
        float GetNewJointRate2(
                const TransitionData::JointTransitionData &jointData,
                AgentBody::JointAxis ax);
        void CheckStability();
        AgentBody::Joint GetJoint(Part bodyPart, int jno);

        virtual void Think() = 0;
};

#endif /*BASEAGENT_H_*/
