/*
 * File: BaseAgentUtility.cpp
 *
 * Description: BaseAgent implementation for utility and old functions
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007,2008
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <ncurses.h>
#include "BaseAgent.h"
#include "LogLevels.h"

#define EC(FUNC)\
do{             \
    echo();     \
    FUNC;       \
    noecho();   \
} while(0)

struct Rates
{
    Rates(): r1(0), r2(0) {}
    float r1, r2;
};

void BaseAgent::UserCommandManager()
{
    int ch;

    printw("User command input activated, to see available commands "
        "enter 'help' or 'h'\n");

    const string statesDir = "data/";
    char tmpstr[100];
    Part selectedPart = LEFT_HAND, oldSelected = LEFT_HAND;
    int jnumber = 1, oldjnum = 1;
    bool updated = false;
    Rates rates[AgentBody::ILLEGAL_JOINT];
    AgentBody::Joint currentJoint = AgentBody::LEFT_ARM_BASE;

    ofstream stateFile((statesDir + "states.st").c_str(), ios::ate | ios::in | ios::out );
    if (!stateFile)
        stateFile.open((statesDir + "states.st").c_str()); // create the file!
    if (!stateFile)
        printw("ERROR: cannot open file states.st for writing states!\n");

    while (wm.IsActive())
    {
        if (selectedPart != oldSelected || jnumber != oldjnum)
        {
            oldSelected = selectedPart;
            oldjnum = jnumber;
            currentJoint = GetJoint(selectedPart, jnumber);
            printw("Current rates: %f %f\n", rates[currentJoint].r1, rates[currentJoint].r2);
        }
        ch = getch();
        switch (ch) {
            case 'h':
            printw("\t* h: this help!\n"
                    "\t* a <act name>: run an action after the current action\n"
                    "\t* a stop: stop current action\n"
                    "\t* w <state name>: save the current agent state in the file\n"
                    "\t* f <file name>: open state file with the given name\n\t"
                    "   for writing. (default: states.st)\n"
                    "\t* o <command>: 'command' will be sent to the server directly\n"
                    "\t* e: select neck\n"
                    "\t* s/d: select left/right hand\n"
                    "\t* z/c: select left/right leg\n"
                    "\t* 1..6: select joint number\n"
                    "\t* Arrow keys: increase/decrease joint velocity\n"
                    "\t* q: quit!\n\n");
                break;
            case 'a':
                printw("action name: ");
                EC(getstr(tmpstr));
                printw("Set as next action: %s\n", tmpstr);
                if (strcmp(tmpstr, "stop") == 0)
                    StopAction();
                else
                    nextAction = acts.ActionType(tmpstr);
                break;
            case 'w':
                printw("state name: ");
                EC(getstr(tmpstr));
                printw("Saving agent state: %s\n", tmpstr);
                wm.SaveAgentState(stateFile, tmpstr);
                stateFile.flush();
                break;
            case 'f':
                printw("file name: ");
                EC(getstr(tmpstr));
                stateFile.close();
                stateFile.open((statesDir + tmpstr).c_str(), ios::ate | ios::in | ios::out );
                if (!stateFile)
                    stateFile.open((statesDir + tmpstr).c_str()); // create the file!
                if (!stateFile)
                    printw("ERROR: cannot open file %s for writing states!\n",
                           (statesDir + tmpstr).c_str());
                else
                    printw("State file set to: %s\n", (statesDir + tmpstr).c_str());
                break;
            case 'o':
                printw("command: ");
                EC(getstr(tmpstr));
                connection.Send(tmpstr);
                break;
            case 'q':
                printw("Exiting...\n");
                wm.Stop();
                break;
            case 'e':
                selectedPart = NECK;
                printw("Neck selected\n");
                break;
            case 's':
                selectedPart = LEFT_HAND;
                printw("Left hand selected\n");
                break;
            case 'd':
                selectedPart = RIGHT_HAND;
                printw("Right hand selected\n");
                break;
            case 'z':
                selectedPart = LEFT_LEG;
                printw("Left leg selected\n");
                break;
            case 'c':
                selectedPart = RIGHT_LEG;
                printw("Right leg selected\n");
                break;
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
                jnumber = ch - '0';
                if ((selectedPart == LEFT_HAND || selectedPart == RIGHT_HAND)
                        && jnumber > 4)
                {
                    printw("Illegal joint number!\n");
                    jnumber = 4;
                }
                else if (selectedPart == NECK && jnumber > 2)
                {
                    printw("Illegal joint number!\n");
                    jnumber = 2;
                }
                printw("joint number: %d\n", jnumber);
                break;
            case KEY_LEFT:
                rates[currentJoint].r1+=0.1;
                updated = true;
                break;
            case KEY_RIGHT:
                rates[currentJoint].r1-=0.1;
                updated = true;
                break;
            case KEY_UP:
                rates[currentJoint].r2+=0.1;
                updated = true;
                break;
            case KEY_DOWN:
                rates[currentJoint].r2-=0.1;
                updated = true;
                break;
            default:
                printw("Unknown command entered: %c\n", ch);
                break;
        }

        if (updated)
        {
            updated = false;
            SetJointVelocity(currentJoint, rates[currentJoint].r1, rates[currentJoint].r2);
            printw("(%s %f %f)\n",
                   AgentBody::JointEffector[GetJoint(selectedPart, jnumber)],
                   rates[currentJoint].r1, rates[currentJoint].r2);
        }
    }
}

AgentBody::Joint BaseAgent::GetJoint(Part bodyPart, int jno)
{
    AgentBody::Joint result = AgentBody::ILLEGAL_JOINT;
    switch (bodyPart) {
        case NECK:
            result = AgentBody::NAO_HEAD_NECK;
            break;
        case RIGHT_HAND:
            result = AgentBody::NAO_RIGHT_TORSO_SHOULDER;
            break;
        case LEFT_HAND:
            result = AgentBody::NAO_LEFT_TORSO_SHOULDER;
            break;
        case RIGHT_LEG:
            result = AgentBody::NAO_RIGHT_TORSO_HIP1;
            break;
        case LEFT_LEG:
            result = AgentBody::NAO_LEFT_TORSO_HIP1;
            break;
        default:
            break;
    }
    if (result != AgentBody::ILLEGAL_JOINT)
        result = static_cast<AgentBody::Joint>(result + jno - 1);
    return result;
}
