/*
 * File: ParseTree.h
 *
 * Description: ParseTree class, based on MBZ's parser class
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2006
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef PARSETREE_H
#define PARSETREE_H

#include <cstdlib>
#include <boost/smart_ptr.hpp>
#include <utility>
#include <vector>
#include <string>

using namespace std;

#ifdef PARSER_SAFE
#define GET(idx) data.at(idx)
#else
#define GET(idx) data[idx]
#endif

/**
 * An S-Expression parser. It'll read an string and creates a ParseTree hierarchy
 * from it. A ParseTree object contains data for a set of parenthesis (not nested).
 * For each nested parenthesis, a separate ParseTree object is created. The first
 * word of each parenthesis is it's key, and the remaining data (usually in a
 * nested parenthesis) is it's data. For example, in
 * (tpkey (skey1 sdata1 sdata2) (skey2 (sk22 sk2d1))), tpkey is the toplevel key
 * and the remaining parts are stored as it's data (which is a ParseTree object).
 * In the contained ParseTree object, skey1 and skey2 are it's keys. The data for
 * skey1 is a ParseTree object containing sdata1 & sdata2 as it's keys and the data
 * for skey2 is a ParseTree containing sk22 as it's key and sk2d1 as the data for
 * sk22. (So, sk2d1 is contained in a separate ParseTree object with sk2d1 as key
 * and no data)
 */
class ParseTree
{
    private:
        typedef boost::shared_ptr<ParseTree> ParseTreePtr;
        typedef vector<pair<string, ParseTreePtr> > DataArray;

    public:
        /** Creates a ParseTree hierarchy from message */
        ParseTree( const string &message ): isOk(true)
        {
            string::size_type pos = 0;
            Parse( message, pos );
        }

        /** Loads file info in a string object removing all comments. Comments
         * start with a '/' and continue to the end of the line
         */
        static string LoadFileToString(const char *filename);
        unsigned GetNumOfKeys() const			{ return data.size(); }

        // Methods for reading key values
        string GetKey( int idx ) const			{ return GET(idx).first; }
        void GetKey( int idx, string &s ) const	{ s = GET(idx).first; }
        void GetKey( int idx, int &i ) const	{ i = atoi( GET(idx).first.c_str() ); }
        void GetKey( int idx, float &f ) const	{ f = atof( GET(idx).first.c_str() ); }
        void GetKey( int idx, double &f ) const	{ f = atof( GET(idx).first.c_str() ); }

        ParseTree *GetData( int idx ) const		{ return GET(idx).second.get(); }

        /** Determines if there were any errors while parsing the message */
        bool IsOk() const						{ return isOk; }

        /** Reads key values in a vector */
        template <typename vecType>
        void GetVector( vecType &v )
        {
            for (int i = 0; i < 3; ++i) {
                GetKey(i, v[i]);
            }
        }

        /** Prints the in memory tree on screen */
        void Show(string indent = "");

        ~ParseTree()
        {}

    private:
        bool isOk;            //< Shows if partitioning has done succefully
        DataArray data;       //< An array of key-data pairs

        /** Creates a ParseTree from message starting from pos */
        ParseTree( const string &message, string::size_type &pos ): isOk(true)
        {
            Parse( message, pos );
        }

        /** Parses the message starting from pos and creates the hierarchy */
        void Parse( const string &message, string::size_type &pos );
};

#endif
