/*
 * File: PolarVector.h
 *
 * Description: Vector Classes
 *
 * Author: 3D Robocup Soccer Team 83 <persisgroup@aim.com>, (c) 2006, 2008
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef POLARVECTOR_H_
#define POLARVECTOR_H_

#include "Vector.h"

class PolarVector3D
{
    private:
        float length;
        AngDeg theta;   // The horizontal angle
        AngDeg phi;     // The latitudinal angle

    public:
        PolarVector3D() :
            length( UnknownFloat ), theta( UnknownAngle ), phi( UnknownAngle )
        {
        }

        explicit PolarVector3D( float length, AngDeg theta = 0, AngDeg phi = 0 )
                : length( length ), theta( theta ), phi( phi )
        {}

        PolarVector3D( const Vector3f &v )
                : length( v.Length() ), theta( atan2Deg( v.y(), v.x() ) ),
                  phi( atan2Deg( v.z(), sqrt( sqr( v.x() ) + sqr( v.y() ) ) ) )
        {}

        PolarVector3D &Set( float length, AngDeg theta, AngDeg phi = 0 )
        {
            this->length = length;
            this->theta = theta;
            this->phi = phi;
            return *this;
        }

        operator Vector3D() const
        {
            return Vector3D( length * cosDeg( theta ) * cosDeg( phi ),
                             length * sinDeg( theta ) * cosDeg( phi ),
                             length * sinDeg( phi )
                           );
        }

        float Length() const
        {
            return length;
        }

        float &Length()
        {
            return length;
        }

        AngDeg Theta() const
        {
            return theta;
        }

        AngDeg &Theta()
        {
            return theta;
        }

        AngDeg Phi() const
        {
            return phi;
        }

        AngDeg &Phi()
        {
            return phi;
        }

        PolarVector3D Normalized() const
        {
            return PolarVector3D( 1, theta, phi );
        }

        PolarVector3D &SetLength( float len )
        {
            length = len;
            return *this;
        }

        PolarVector3D operator+( const PolarVector3D &vec ) const
        {
            return Vector3D(*this) + vec;
        }

        PolarVector3D operator-( const PolarVector3D &vec ) const
        {
            return Vector3D(*this) - vec;
        }

        float Length2D() const
        {
            return Vector3D(*this).To2D().Length();
        }
};

const PolarVector3D UnknownPolarVector;

inline std::ostream &operator<<( std::ostream &s, const PolarVector3D &vec )
{
    s << "pol( " << vec.Length() << ", " << vec.Theta() << ", " << vec.Phi() << " )";
    return s;
}

#endif /*POLARVECTOR_H_*/
