/*
 * File: TransitionData.cpp
 *
 * Description: TransitionData class implementation
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef PARSER_SAFE
#define PARSER_SAFE
#endif

#include "TransitionData.h"
#include "ParseTree.h"
#include "Exception.h"
#include <algorithm>
using namespace std;

void TransitionData::AddJointAction(const JointTransitionData &jointData)
{
    ThrowIf(jointData.rate[0]<0.0||jointData.rate[1]<0.0,
            "The rate values in action files should be positive!");
}

void TransitionData::ReadTransitionData( ParseTree *transition )
{
    for (unsigned i = 0; i < transition->GetNumOfKeys(); ++i)
    {
        if (transition->GetKey(i) == "SetParam")
        {
            string name = transition->GetData(i)->GetKey(0);
            float value;
            transition->GetData(i)->GetKey(1, value);
            Log(LOG_TRANSITION_DATA) << "Setting parameter " << name << " to " << value;
            parameters[name] = value;
        }
        else if (transition->GetKey(i) == "Step")
        {
            ParseTree *step = transition->GetData(i);
            if (step->GetKey(0) != "name")
            {
                Log(LOG_ERROR) << "\tFirst data must be the step's name!";
                return;
            }
            Log(LOG_TRANSITION_DATA) << "\tLoading step: " << step->GetData(0)->GetKey(0);
            steps.push_back(StepData(step->GetData(0)->GetKey(0)));

            bool jointAdded[AgentBody::ILLEGAL_JOINT];
            fill(jointAdded, jointAdded+AgentBody::ILLEGAL_JOINT, false);
            for (unsigned i = 1; i < step->GetNumOfKeys(); ++i)
            {
                if (step->GetKey(i) == "SetAll")
                {
                    if (step->GetData(i)->GetKey(0) == "false")
                    {
                        steps.back().setAllJoints = false;
                        Log(LOG_TRANSITION_DATA) << "\tDon't Set all joint rates.";
                    }
                    else if (step->GetData(i)->GetKey(0) == "true")
                    {
                        steps.back().setAllJoints = true;
                        Log(LOG_TRANSITION_DATA) << "\tSet all joint rates.";
                    }
                }
                else if (step->GetKey(i) == "Call")
                {
                    Log(LOG_TRANSITION_DATA) << "\tSet ruby call: "
                            << step->GetData(i)->GetKey(0);
                    steps.back().checkFunction = step->GetData(i)->GetKey(0);
                }
                else
                {
                    JointTransitionData jointData;
                    Log(LOG_TRANSITION_DATA) << "\tLoading joint info: "
                            << step->GetKey(i);
                    jointData.joint = AgentBody::joint(step->GetKey(i));
                    jointData.ReadJointData(step->GetData(i));
                    jointAdded[jointData.joint] = true;
                    steps.back().actionList.push_back(jointData);
                }
            }

            if (steps.back().setAllJoints)
            {
                for (int i = 0; i < AgentBody::CurrentFinalJoint; ++i)
                    if (!jointAdded[i])
                    {
                        Log(LOG_TRANSITION_DATA) << "\tSet zero rates for: "
                                << AgentBody::JointName[i];
                        JointTransitionData jointData;
                        jointData.joint = static_cast<AgentBody::Joint>(i);
                        jointData.rate[0] = jointData.rate[1] = jointData.targetax[0]
                                = jointData.targetax[1] = 0;
                        steps.back().actionList.push_back(jointData);
                    }
            }
        }
        else if (transition->GetKey(i) != "From"&& transition->GetKey(i) != "To")
            Log(LOG_TRANSITION_DATA) << "Unknown tag in transition: "
                    << transition->GetKey(i);
    }
}

inline void TransitionData::JointTransitionData::ReadParam(string str, float &val, string &param)
{
    if (str[0] == '$')
        param = str;
    else
        val = atof(str.c_str());
}

void TransitionData::JointTransitionData::ReadJointData(ParseTree *jointInfo)
{
    if (joint==AgentBody::ILLEGAL_JOINT)
    {
        ReadParam(jointInfo->GetKey(0), delay, rateparam[0]);
        Log(LOG_TRANSITION_DATA) << "\t\tSet time delay: " << delay << ' ' << rateparam[0];
    }
    else
    {
        for (unsigned i = 0; i < jointInfo->GetNumOfKeys(); ++i)
        {
            if (jointInfo->GetKey(i) == "rates")
            {
                ReadParam(jointInfo->GetData(i)->GetKey(0), rate[0], rateparam[0]);
                Log(LOG_TRANSITION_DATA) << "\t\tSet rate[0]: " << rate[0] << ' ' << rateparam[0];
                if (jointInfo->GetData(i)->GetNumOfKeys() > 1)
                {
                    ReadParam(jointInfo->GetData(i)->GetKey(1), rate[1], rateparam[1]);
                    Log(LOG_TRANSITION_DATA) << "\t\tSet rate[1]: " << rate[1] << ' ' << rateparam[1];
                }
            }
            else if (jointInfo->GetKey(i) == "target")
            {
                ReadParam(jointInfo->GetData(i)->GetKey(0), targetax[0], targetaxparam[0]);
                Log(LOG_TRANSITION_DATA) << "\t\tSet targetax[0]: " << targetax[0] << ' ' << targetaxparam[0];
                if (jointInfo->GetData(i)->GetNumOfKeys() > 1)
                {
                    ReadParam(jointInfo->GetData(i)->GetKey(1), targetax[1], targetaxparam[1]);
                    Log(LOG_TRANSITION_DATA) << "\t\tSet targetax[1]: " << targetax[1] << ' ' << targetaxparam[1];
                }
            }
            else
                Log(LOG_ERROR) << "Unknown tag in joint data: " << jointInfo->GetKey(i);
        }
    }
}

void TransitionData::JointTransitionData::UpdateValues(const TransitionData *actions)
{
    if (rateparam[0].size())
        rate[0] = actions->GetParameter(rateparam[0]);
    if (rateparam[1].size())
        rate[1] = actions->GetParameter(rateparam[1]);
    if (targetaxparam[0].size())
        targetax[0] = actions->GetParameter(targetaxparam[0]);
    if (targetaxparam[1].size())
        targetax[1] = actions->GetParameter(targetaxparam[1]);
}
