/*
 * File: Vector.h
 *
 * Description: Vector Classes
 *
 * Author: UI-AI 3D Robocup Soccer Team 83 <persisgroup@aim.com>, (c) 2006
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef VECTOR_H_
#define VECTOR_H_

#include <ostream>
#include <salt/vector.h>
#include <salt/matrix.h>
#include "undef_package.h"
#include "MathUtils.h"

using namespace salt;

class Vector2D: public Vector2f
{
    public:
        Vector2D(): Vector2f( UnknownAngle, UnknownAngle )
        {}

        Vector2D( const Vector2f &v ): Vector2f( v )
        {}

        Vector2D( float x, float y ): Vector2f( x, y )
        {}

        const Vector2D &Set( float x, float y )
        {
            return static_cast<const Vector2D &>( Vector2f::Set( x, y ) );
        }

        double GetDistanceTo( Vector2D v ) const
        {
            return ( v - *this ).Length();
        }

        Vector2D &SetLength( float len )
        {
            Normalize();
            ( *this ) *= len;
            return *this;
        }

        bool IsBetweenX( float x1, float x2 ) const
        {
            return x() > x1 && x() < x2;
        }

        bool IsBetweenY( float y1, float y2 ) const
        {
            return y() > y1 && y() < y2;
        }

        bool IsLeftOF( Vector2D Vec ) const
        {
            return( Vec.x() * y()  - x() * Vec.y()  > 0 );

        }

        bool IsRightOF( Vector2D Vec ) const
        {
            return( Vec.x() * y()  - x() * Vec.y()  < 0 );
        }

        bool IsZero() const
        {
            return x() < std::numeric_limits<float>::epsilon() &&
                   y() < std::numeric_limits<float>::epsilon();
        }

        bool OnTheSameLine( Vector2D Vec ) const
        {
            return( Vec.x() * y()  - x() * Vec.y()  == 0.0 );
        }

        /**
         * int side =
         * 1 both on right side
         * 2 both on left side
         * 3 *this on left , vec2 on right
         * 4 *this on right , vec2 on left
         **/
        void SideRelativeTo( Vector2D RelVec, Vector2D Vec2 , int &side )
        {
            if ( ( *this ).IsRightOF( RelVec ) && Vec2.IsRightOF( RelVec ) )
                side = 1;
            if ( ( *this ).IsLeftOF( RelVec ) && Vec2.IsLeftOF( RelVec ) )
                side = 2;

            if ( ( *this ).IsLeftOF( RelVec ) && Vec2.IsRightOF( RelVec ) )
                side = 3;
            if ( ( *this ).IsRightOF( RelVec ) && Vec2.IsLeftOF( RelVec ) )
                side = 4;

            return;
        }
};

class Vector3D: public Vector3f
{
    public:
        Vector3D(): Vector3f( UnknownAngle, UnknownAngle, UnknownAngle )
        {}

        Vector3D( const Vector3f &v ): Vector3f( v )
        {}

        Vector3D( const Vector2f &v ): Vector3f( v.x(), v.y(), 0 )
        {}

        Vector3D( float x, float y, float z = 0 ): Vector3f( x, y, z )
        {}

        const Vector3D &Set( float x, float y, float z = 0 )
        {
            return static_cast<const Vector3D &>( Vector3f::Set( x, y, z ) );
        }

        Vector2D To2D() const
        {
            return Vector2D(x(), y());
        }

        operator Vector2D() const
        {
            return Vector2D(x(), y());
        }

        double GetDistanceTo( Vector3D v ) const
        {
            return ( v - *this ).Length();
        }

        Vector3D &SetLength( float len )
        {
            Normalize();
            ( *this ) *= len;
            return *this;
        }

        Vector3D &RotateXY( AngDeg angle )
        {
            Matrix m;
            m.RotationZ( angle );
            *this = m.Transform( *this );
            return *this;
        }

        bool IsBetweenX( float x1, float x2 ) const
        {
            return x() > x1 && x() < x2;
        }

        bool IsBetweenY( float y1, float y2 ) const
        {
            return y() > y1 && y() < y2;
        }

        bool IsBetweenZ( float z1, float z2 ) const
        {
            return z() > z1 && z() < z2;
        }

        bool IsZero() const
        {
            return x() < std::numeric_limits<float>::epsilon() &&
                   y() < std::numeric_limits<float>::epsilon() &&
                   z() < std::numeric_limits<float>::epsilon();
        }

};

const Vector3D UnknownVector3D;
const Vector2D UnknownVector2D;

inline std::ostream &operator<<( std::ostream &s, const Vector3D &vec )
{
    s << "xyz( " << vec.x() << ", " << vec.y() << ", " << vec.z() << " )";
    return s;
}

#endif /*VECTOR_H_*/
