/*
 * File: WorldModelCore.cpp
 *
 * Description: WorldModelCore implementation file
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "WorldModelCore.h"
#include "LogLevels.h"
#include "ParseTree.h"
#include "ConnectionManager.h"
#include <rcssserver3d/soccer/soccertypes.h>

#ifdef NO_INLINE
#include "WorldModelCore.inline.h"
#endif

using namespace std;
using namespace boost;

WorldModelCore::WorldModelCore(ConnectionManager *connection) :
    running(false), teamName("UIAI"), connection(connection), simtime(0),
            deltaTime(0), gameTime(0), newSenseInfo(false)
{
    SetupVisionPerceptorMap(false);
    SetupPlayModeMap(false);

    flagPosition[FLAG_LEFT_US] = Vector3D(FieldLength() / -2.0, FieldWidth() / +2.0, 0.0);
    flagPosition[FLAG_RIGHT_US] = Vector3D(FieldLength() / -2.0, FieldWidth() / -2.0, 0.0);
    flagPosition[FLAG_LEFT_THEM] = Vector3D(FieldLength() / +2.0, FieldWidth() / +2.0, 0.0);
    flagPosition[FLAG_RIGHT_THEM] = Vector3D(FieldLength() / +2.0, FieldWidth() / -2.0, 0.0);

    flagPosition[GOAL_RIGHT_US] = Vector3D(FieldLength() / -2.0, GoalWidth() / -2.0, 0.0);
    flagPosition[GOAL_LEFT_US] = Vector3D(FieldLength() / -2.0, GoalWidth() / +2.0, 0.0);
    flagPosition[GOAL_RIGHT_THEM] = Vector3D(FieldLength() / +2.0, GoalWidth() / -2.0, 0.0);
    flagPosition[GOAL_LEFT_THEM] = Vector3D(FieldLength() / +2.0, GoalWidth() / +2.0, 0.0);
}

void WorldModelCore::SetupVisionPerceptorMap(bool isLeftTeam)
{
    if ( isLeftTeam )
    {
        visionPerceptorMap["F1L"] = FLAG_LEFT_US;
        visionPerceptorMap["F1R"] = FLAG_LEFT_THEM;
        visionPerceptorMap["F2L"] = FLAG_RIGHT_US;
        visionPerceptorMap["F2R"] = FLAG_RIGHT_THEM;
        visionPerceptorMap["G1L"] = GOAL_LEFT_US;
        visionPerceptorMap["G1R"] = GOAL_LEFT_THEM;
        visionPerceptorMap["G2L"] = GOAL_RIGHT_US;
        visionPerceptorMap["G2R"] = GOAL_RIGHT_THEM;
    }
    else
    {
        visionPerceptorMap["F1L"] = FLAG_RIGHT_THEM;
        visionPerceptorMap["F2L"] = FLAG_LEFT_THEM;
        visionPerceptorMap["F1R"] = FLAG_RIGHT_US;
        visionPerceptorMap["F2R"] = FLAG_LEFT_US;
        visionPerceptorMap["G1L"] = GOAL_RIGHT_THEM;
        visionPerceptorMap["G2L"] = GOAL_LEFT_THEM;
        visionPerceptorMap["G1R"] = GOAL_RIGHT_US;
        visionPerceptorMap["G2R"] = GOAL_LEFT_US;
    }
    visionPerceptorMap["B"] = BALL;
    visionPerceptorMap["PT1"] = TEAMMATE_1;
    visionPerceptorMap["PT2"] = TEAMMATE_2;
    visionPerceptorMap["PT3"] = TEAMMATE_3;
    visionPerceptorMap["PT4"] = TEAMMATE_4;
    visionPerceptorMap["PT5"] = TEAMMATE_5;
    visionPerceptorMap["PT6"] = TEAMMATE_6;
    visionPerceptorMap["PT7"] = TEAMMATE_7;
    visionPerceptorMap["PT8"] = TEAMMATE_8;
    visionPerceptorMap["PT9"] = TEAMMATE_9;
    visionPerceptorMap["PT10"] = TEAMMATE_10;
    visionPerceptorMap["PT11"] = TEAMMATE_11;
    visionPerceptorMap["PO1"] = OPPONENT_1;
    visionPerceptorMap["PO2"] = OPPONENT_2;
    visionPerceptorMap["PO3"] = OPPONENT_3;
    visionPerceptorMap["PO4"] = OPPONENT_4;
    visionPerceptorMap["PO5"] = OPPONENT_5;
    visionPerceptorMap["PO6"] = OPPONENT_6;
    visionPerceptorMap["PO7"] = OPPONENT_7;
    visionPerceptorMap["PO8"] = OPPONENT_8;
    visionPerceptorMap["PO9"] = OPPONENT_9;
    visionPerceptorMap["PO10"] = OPPONENT_10;
    visionPerceptorMap["PO11"] = OPPONENT_11;
}

void WorldModelCore::SetupPlayModeMap(bool isLeftTeam)
{
    playModeMap[STR_PM_BeforeKickOff    ] = BEFORE_KICKOFF;
    playModeMap[STR_PM_PlayOn           ] = PLAYON;
    playModeMap[STR_PM_GameOver         ] = GAME_OVER;
    playModeMap[STR_PM_Unknown          ] = UNKNOWN_PLAYMODE;
    if ( isLeftTeam )
    {
        playModeMap[STR_PM_KickOff_Left     ] = KICKOFF_US;
        playModeMap[STR_PM_KickOff_Right    ] = KICKOFF_THEM;
        playModeMap[STR_PM_KickIn_Left      ] = KICKIN_US;
        playModeMap[STR_PM_KickIn_Right     ] = KICKIN_THEM;
        playModeMap[STR_PM_CORNER_KICK_LEFT ] = CORNER_KICK_US;
        playModeMap[STR_PM_CORNER_KICK_RIGHT] = CORNER_KICK_THEM;
        playModeMap[STR_PM_GOAL_KICK_LEFT   ] = GOAL_KICK_US;
        playModeMap[STR_PM_GOAL_KICK_RIGHT  ] = GOAL_KICK_THEM;
        playModeMap[STR_PM_OFFSIDE_LEFT     ] = OFFSIDE_US;
        playModeMap[STR_PM_OFFSIDE_RIGHT    ] = OFFSIDE_THEM;
        playModeMap[STR_PM_Goal_Left        ] = GOAL_US;
        playModeMap[STR_PM_Goal_Right       ] = GOAL_THEM;
        playModeMap[STR_PM_FREE_KICK_LEFT   ] = FREE_KICK_US;
        playModeMap[STR_PM_FREE_KICK_RIGHT  ] = FREE_KICK_THEM;
    }
    else
    {
        playModeMap[STR_PM_KickOff_Left     ] = KICKOFF_THEM;
        playModeMap[STR_PM_KickOff_Right    ] = KICKOFF_US;
        playModeMap[STR_PM_KickIn_Left      ] = KICKIN_THEM;
        playModeMap[STR_PM_KickIn_Right     ] = KICKIN_US;
        playModeMap[STR_PM_CORNER_KICK_LEFT ] = CORNER_KICK_THEM;
        playModeMap[STR_PM_CORNER_KICK_RIGHT] = CORNER_KICK_US;
        playModeMap[STR_PM_GOAL_KICK_LEFT   ] = GOAL_KICK_THEM;
        playModeMap[STR_PM_GOAL_KICK_RIGHT  ] = GOAL_KICK_US;
        playModeMap[STR_PM_OFFSIDE_LEFT     ] = OFFSIDE_THEM;
        playModeMap[STR_PM_OFFSIDE_RIGHT    ] = OFFSIDE_US;
        playModeMap[STR_PM_Goal_Left        ] = GOAL_THEM;
        playModeMap[STR_PM_Goal_Right       ] = GOAL_US;
        playModeMap[STR_PM_FREE_KICK_LEFT   ] = FREE_KICK_THEM;
        playModeMap[STR_PM_FREE_KICK_RIGHT  ] = FREE_KICK_US;
    }
}

void WorldModelCore::Sense()
{
    try
    {
        running = true;
        while(running)
        {
            Log(LOG_DEBUG)<<"Reading wait...";
            string msg = connection->Receive();
            ParseTree parser(msg);
            if (!parser.IsOk() || msg.size() == 0)
                continue;
            bool gonext = true;
            for (unsigned i = 0; gonext && i < parser.GetNumOfKeys(); ++i)
                gonext = SelectParser(parser.GetKey(i), parser.GetData(i));
            Log(LOG_DEBUG)<<"Notify";
            newSense.notify_all();
        }
    }
    catch (std::exception &e)
    {
        Log(LOG_ERROR) << "Oh! Something happened in Sense: " << e.what();
    }
    catch (...)
    {
        Log(LOG_ERROR, "Oh! Something happened in Sense?!!");
    }
    running=false;
    newSense.notify_all();
}

void WorldModelCore::WaitForNewSense()
{
    mutex::scoped_lock lock(waitMutex);
    newSense.wait(lock);
}

inline void WorldModelCore::SetTime(float tm)
{
    if (tm>simtime)
    {
        deltaTime = tm - simtime;
        simtime = tm;
        newSenseInfo = true;
    }
}

inline void WorldModelCore::SetVisionObjectInfo(string flagName, PolarVector3D p)
{
    Log(LOG_DEBUG_SENSE) << flagName << " -> " << p
                << " Vector3D: " << Vector3D(p);

    VisionObjectMap::const_iterator flagptr = visionPerceptorMap.find(flagName);
    if (flagptr == visionPerceptorMap.end())
        Log( LOG_ERROR ) << "Unknown flag percept received: " << flagName;
    else
        visionObjectsPolar[ flagptr->second ] = p;
}

void WorldModelCore::UpdateMyPositionAndVelocity()
{
    MyOldPosition = MyPosition;

    PolarVector3D flag1 = visionObjectsPolar[FLAG_LEFT_US];
    PolarVector3D flag2 = visionObjectsPolar[FLAG_RIGHT_US];
    PolarVector3D flag3 = visionObjectsPolar[FLAG_LEFT_THEM];
    const Vector3D& f1 = flagPosition[FLAG_LEFT_US];
    const Vector3D& f2 = flagPosition[FLAG_RIGHT_US];
    const Vector3D& f3 = flagPosition[FLAG_LEFT_THEM];
    float x2 = f2.x() - f1.x();
    float x3 = f3.x() - f1.x();
    float y2 = f2.y() - f1.y();
    float y3 = f3.y() - f1.y();
    float t2 = sqr(x2)+sqr(y2)-sqr(flag2.Length())+sqr(flag1.Length());
    float t3 = sqr(x3)+sqr(y3)-sqr(flag3.Length())+sqr(flag1.Length());
    float y = (t3*x2-t2*x3) / (2*(y3*x2 - y2*x3));
    float x = -( 2 * y * y3 - t3 ) / ( 2 * x3 );
    float z2 = sqr(flag1.Length()) - sqr(x) - sqr(y);
    float z = 0;
    if (z2 >= 0)
        z = sqrt(z2);
    MyPosition.x() = x + f1.x();
    MyPosition.y() = y + f1.y();
    MyPosition.z() = z + f1.z();

    CalculateRotationValues();

    MyVelocity = (MyPosition - MyOldPosition) / deltaTime;
    Log(LOG_VISION) << "My position: " << MyPosition;
}

void WorldModelCore::UpdateBallPositionAndVelocity()
{
    BallOldPosition = BallPosition;
    BallPosition = GetRelativePolarVectorTo( BALL);
    BallVelocity = (BallPosition - BallOldPosition) / deltaTime;
    Log(LOG_VISION) << "Ball Velocity: " << BallVelocity;
}


bool WorldModelCore::SelectParser(string key, ParseTree *parser)
{
    if (key=="time")
        return ParseTime(parser);
    else if (key == "GS")
        ParseGameState(parser);
    else if (key == "See")
        ParseVision(parser);
    else if (key=="UJ")
        ParseUniversalJointInfo(parser);
    else if (key=="HJ")
        ParseHingeJointInfo(parser);
    else if (key=="GYR")
        ParseGyroInfo(parser);
//	else if (key=="TCH")
//		ParseTouchInfo(parser);
    else if (key=="FRP")
        ParseForceResistanceInfo(parser);
    else
        Log(LOG_WARNING) << "Unknown sense received: " << key;
    return true;
}

bool WorldModelCore::ParseTime(ParseTree *parser)
{
    float time = 0;
    for (unsigned i = 0; i < parser->GetNumOfKeys(); ++i) {
        if (parser->GetKey(i)=="now")
            parser->GetData(i)->GetKey(0,time);
    }
    bool res = Time() < time;
    if (res)
    {
        SetTime(time);
        // NOTICE: It assumes that we always get time percept before FRP percept
        agent.ClearFRPData();
    }
    Log(LOG_DEBUG_SENSE) << "Time: " << time;

    return res;
}

void WorldModelCore::ParseGameState(ParseTree *parser)
{
    for (unsigned i = 0; i < parser->GetNumOfKeys(); ++i)
    {
        if (parser->GetKey(i)=="t")
            parser->GetData(i)->GetKey(0, gameTime);
        else if (parser->GetKey(i)=="pm")
        {
            string playModeStr = parser->GetData(i)->GetKey(0);
            PlayModeMap::const_iterator playModePtr = playModeMap.find(playModeStr);
            if (playModePtr != playModeMap.end())
                playMode = playModePtr->second;
            else
                Log(LOG_WARNING) << "Unknown play mode: "<< playModeStr;
        }
        else if (parser->GetKey(i)=="unum")
        {
            parser->GetData(i)->GetKey(0, playerNumber);
/*			char filename[100];
            sprintf(filename, "logs/log%d.log", playerNumber);
            logOutput.open(filename);
            Log.setOutputStream(logOutput);*/
        }
        else if (parser->GetKey(i)=="team")
        {
            bool isLeftTeam = parser->GetData(i)->GetKey(0) == "left" ? true : false;
            if (isLeftTeam)
            {
                SetupVisionPerceptorMap(isLeftTeam);
                SetupPlayModeMap(isLeftTeam);
            }
        }
    }
}

void WorldModelCore::ParseVision(ParseTree *parser)
{
    string objectName;
    float r = UnknownFloat, t = UnknownAngle, p = UnknownAngle;
    for (unsigned i = 0; i < parser->GetNumOfKeys(); ++i)
    {
        objectName = parser->GetKey(i);
        ParseTree*visionInfo = parser->GetData(i);

        for (unsigned j = 0; j < visionInfo->GetNumOfKeys(); ++j)
        {
            if (visionInfo->GetKey(j) == "team")
                if (visionInfo->GetData(j)->GetKey(0) == teamName)
                    objectName += "T";
                else
                    objectName += "O";
            else if (visionInfo->GetKey(j) == "id")
                objectName += visionInfo->GetData(j)->GetKey(0);
            else if (visionInfo->GetKey(j) == "pol")
            {
                visionInfo->GetData(j)->GetKey(0, r);
                visionInfo->GetData(j)->GetKey(1, t);
                visionInfo->GetData(j)->GetKey(2, p);
            }
        }

        SetVisionObjectInfo( objectName, PolarVector3D(r,t,p) );
    }
    UpdateMyPositionAndVelocity();
    UpdateBallPositionAndVelocity();

}

void WorldModelCore::ParseHingeJointInfo(ParseTree *parser)
{
    string jointName;
    float axis1 = UnknownAngle;
    for (unsigned i = 0; i < parser->GetNumOfKeys(); ++i)
    {
        if (parser->GetKey(i)=="n")
            parser->GetData(i)->GetKey(0, jointName);
        else if (parser->GetKey(i)=="ax")
            parser->GetData(i)->GetKey(0, axis1);
    }
    agent.SetJointInfo(jointName, deltaTime, axis1);
}

void WorldModelCore::ParseUniversalJointInfo(ParseTree *parser)
{
    string jointName;
    float axis1 = UnknownAngle, axis2 = UnknownAngle;
    for (unsigned i = 0; i < parser->GetNumOfKeys(); ++i)
    {
        if (parser->GetKey(i)=="n")
            parser->GetData(i)->GetKey(0, jointName);
        else if (parser->GetKey(i)=="ax1")
            parser->GetData(i)->GetKey(0, axis1);
        else if (parser->GetKey(i)=="ax2")
            parser->GetData(i)->GetKey(0, axis2);
    }
    agent.SetJointInfo(jointName, deltaTime, axis1, axis2);
}

void WorldModelCore::ParseGyroInfo(ParseTree *parser)
{
    string gyrName;
    Vector3D gyroRate;
    for (unsigned i = 0; i < parser->GetNumOfKeys(); ++i)
    {
        if (parser->GetKey(i)=="n")
            parser->GetData(i)->GetKey(0, gyrName);
        else if (parser->GetKey(i)=="rt")
            parser->GetData(i)->GetVector(gyroRate);
    }
    agent.SetGyroRate(gyrName, gyroRate);
}

void WorldModelCore::ParseTouchInfo(ParseTree *parser)
{
    string tchName;
    int val;
    for (unsigned i = 0; i < parser->GetNumOfKeys(); ++i)
    {
        if (parser->GetKey(i)=="n")
            parser->GetData(i)->GetKey(0, tchName);
        else if (parser->GetKey(i)=="val")
            parser->GetData(i)->GetKey(0, val);
    }
    // update worldmodel with touch info (currently unused!)
}

void WorldModelCore::ParseForceResistanceInfo(ParseTree *parser)
{
    string frpName;
    Vector3D center, force;
    for (unsigned i = 0; i < parser->GetNumOfKeys(); ++i)
    {
        if (parser->GetKey(i)=="n")
            parser->GetData(i)->GetKey(0, frpName);
        else if (parser->GetKey(i)=="c")
            parser->GetData(i)->GetVector(center);
        else if (parser->GetKey(i)=="f")
            parser->GetData(i)->GetVector(force);
    }
    agent.SetForceResistanceInfo(frpName, center, force);
}

void WorldModelCore::CalculateRotationValues()
{
    PolarVector3D current = static_cast<Vector3D>(visionObjectsPolar[FLAG_RIGHT_THEM]
            + visionObjectsPolar[FLAG_LEFT_THEM])/2.0;

//	Log(LOG_VISION) << "Current vector is " << current;

    zRotation.RotationZ(gDegToRad(-current.Theta()));
    yRotation.RotationY(gDegToRad(current.Phi()));

//	PolarVector centerRot = yRotation.Rotate(zRotation.Rotate(current));
//	Log(LOG_VISION) << "Curren after rotation: " << centerRot;

    Vector3D leftflag = yRotation.Rotate(zRotation.Rotate(visionObjectsPolar[FLAG_LEFT_THEM]));
    xRotation.RotationX(-atan2(leftflag.z(),leftflag.y()));

    PolarVector3D targetc = Vector3D(FieldLength()/2, 0, 0) - MyPosition;
    PolarVector3D targetf = flagPosition[FLAG_LEFT_THEM] - MyPosition;
    zInvRotation.RotationZ(-gDegToRad(targetc.Theta()));
    yInvRotation.RotationY(gDegToRad(targetc.Phi()));

    Vector3D tf = yInvRotation.Rotate(zInvRotation.Rotate(targetf));
    xInvRotation.RotationX(-atan2(tf.z(),tf.y()));
}
