/*
 * File: WorldModelCore.h
 *
 * Description: Provides the worldmodel data store. Only the SenseHandler could
 * 				update the information externally.
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef WORLDMODELCORE_H_
#define WORLDMODELCORE_H_

#include <string>
#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>
#include <fstream>
#include "AgentBody.h"
#include "Vector.h"
#include "PolarVector.h"
#include "SoccerTypes.h"

using namespace std;
using namespace salt;

class ParseTree;
class ConnectionManager;

/**
 * This is the basic class which provides sense information
 */
 class WorldModelCore
{
    public:
        WorldModelCore(ConnectionManager *connection);
        virtual ~WorldModelCore() {}

        void Sense();
        void Stop();
        void WaitForNewSense();
        bool IsActive() const;
        bool IsNewSenseInfo();
        float Time() const;
        float GameTime() const;
        string TeamName() const;
        Vector3D MyPos() const;
        Vector3D BallVel() const;
        PlayModeType PlayMode() const;
        int PlayerNumber() const;
        float GetSpentTime( float start ) const;
        AngDeg JointAxis(AgentBody::Joint joint, AgentBody::JointAxis ax) const;
        AgentBody::JointInfo JointInformation(AgentBody::Joint joint) const;
        const PolarVector3D &GetRelativePolarVectorTo(VisionObject obj) const;
        PolarVector3D GetRelativePolarVectorTo(const Vector3D &globalPosition) const;
        PolarVector3D GetGlobalRelativePolarVectorTo(VisionObject obj) const;
        PolarVector3D GetGlobalRelativePolarVectorTo(const Vector3D &globalPosition) const;
        const Vector3D &GyroRate() const;
        const AgentBody::ForceResistanceInfo
                &FootTouchForceInfo(AgentBody::AgentFoot foot) const;
        bool IsInGround() const;
        void SetTeamName(const string &teamName);
        void SaveAgentState(ostream &os, const string &stateName);
        Vector3D GetGlobalPosition(VisionObject obj) const;

        // field information
        float FieldWidth() const;
        float FieldLength() const;
        float GoalWidth() const;

    private:
        typedef map<string, VisionObject> VisionObjectMap;
        typedef map<string, PlayModeType> PlayModeMap;

        bool running;
        string teamName;
        ConnectionManager *connection;
        boost::mutex waitMutex;
        boost::condition newSense;
        int playerNumber;
        float simtime, deltaTime;
        float gameTime;
        bool newSenseInfo;
        AgentBody agent;
        PlayModeType playMode;
        Vector3D MyPosition, MyOldPosition, MyVelocity;
        Vector3D BallPosition,BallOldPosition,BallVelocity;
        VisionObjectMap visionPerceptorMap;
        PlayModeMap playModeMap;
        PolarVector3D visionObjectsPolar[ILLEGAL_OBJECT];
        Vector3D flagPosition[BALL];
        Matrix xRotation;
        Matrix yRotation;
        Matrix zRotation;
        Matrix xInvRotation;     // TODO: This MUST be merged with xRotation
        Matrix yInvRotation;
        Matrix zInvRotation;

        void SetTime(float tm);
        void SetupVisionPerceptorMap(bool isLeftTeam);
        void SetupPlayModeMap(bool isLeftTeam);
        void SetVisionObjectInfo(string flagName, PolarVector3D p);
        void UpdateMyPositionAndVelocity(void);
        void UpdateBallPositionAndVelocity();
        bool SelectParser(std::string key, ParseTree *parser);
        bool ParseTime(ParseTree *parser);
        void ParseGameState(ParseTree *parser);
        void ParseVision(ParseTree *parser);
        void ParseHingeJointInfo(ParseTree *parser);
        void ParseUniversalJointInfo(ParseTree *parser);
        void ParseGyroInfo(ParseTree *parser);
        void ParseTouchInfo(ParseTree *parser);
        void ParseForceResistanceInfo(ParseTree *parser);
        void CalculateRotationValues();
        Vector3D Rotate(const Vector3D &relative) const;
};

#ifndef NO_INLINE
#include "WorldModelCore.inline.h"
#endif

#endif /*WORLDMODELCORE_H_*/
